@@ -695,7 +695,8 @@ ecc_generate (const gcry_sexp_t genparms, gcry_sexp_t *r_skey)
 
   if (fips_mode () && test_keys_fips (*r_skey))
     {
-      sexp_release (*r_skey); r_skey = NULL;
+      sexp_release (*r_skey);
+      *r_skey = NULL;
       fips_signal_error ("self-test after key generation failed");
       rc = GPG_ERR_SELFTEST_FAILED;
     }
